import functools


def call_increase(value):
    """
    decrease value when func called.
    :param value: integer
    :return: decorator
    """

    value += 1

    def wrapper(func):
        @functools.wraps(func)
        def inner_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return inner_wrapper

    return wrapper


def call_decrease(value):
    """
    decrease value when func called.
    :param value: integer
    :return: decorator
    """

    value -= 1

    def wrapper(func):
        @functools.wraps(func)
        def inner_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return inner_wrapper

    return wrapper



from typing import Callable
from tornado import httpclient
from tornado.httpclient import HTTPRequest
from ..downloader.base import BaseDownloader
from .. import setting


class AsyncDownloader(BaseDownloader):
    """
    parallel downloader.
    """

    def __init__(self):
        self.client = httpclient.AsyncHTTPClient(max_clients=setting.max_concurrency)

    def fetch(self, request: HTTPRequest, callback: Callable) -> None:
        """
        fetch request and return response
        """

        self.client.fetch(request, callback)



import time
from typing import Callable, Tuple
from tornado.httpclient import HTTPRequest
from ..downloader.base import BaseDownloader
from ..utils.call import call_increase, call_decrease
from .. import setting


class BaseScheduler(object):
    """
    Interface for Scheduler class
    """

    def __init__(self, downloader: BaseDownloader):
        self.downloader = downloader
        self.max_concurrency = setting.max_concurrency
        self.concurrency = 0

        # increase concurrency when fetch func called
        flag_string = 'cb_decrease_flag'
        if not getattr(self.downloader.fetch, flag_string, False):
            self.downloader.fetch = call_increase(self.concurrency)(self.downloader.fetch)
            setattr(self.downloader.fetch, flag_string, True)

    def get(self) -> Tuple[HTTPRequest, Callable]:
        """
        get request obj and callback func from scheduler
        :return: tuple of HTTPRequest and Callable func or None
        """
        pass

    def put(self, request: HTTPRequest, callback: Callable) -> None:
        """
        put request item from scheduler
        :param request:
        :param callback: call back func
        :return: None
        """
        flag_string = 'cb_decrease_flag'
        if not getattr(callback, flag_string, False):
            callback = call_decrease(self.concurrency)(callback)
            setattr(callback, flag_string, True)

        while not self.empty() and self.concurrency < self.max_concurrency:
            ret = self.get()
            if ret:
                req, cb = ret
                self.downloader.fetch(req, cb)

    def empty(self) -> bool:
        """
        scheduler's queue empty or not
        :return: Boolean, True or False
        """
        pass



from typing import Callable
from tornado.httpclient import HTTPRequest


class BaseDownloader(object):
    """
    Interface for Downloader
    """

    def fetch(self, request: HTTPRequest, callback: Callable) -> bool:
        """
        Fetch page, and call callback.
        Return True is request start, else return False.
        :param request: HTTPRequest object
        :param callback: func with HTTPResponse object as arguments
        :return: boolean, start fetch current page or not.
        """
        pass



from tornado import httpclient
from tornado.httpclient import HTTPRequest, HTTPResponse
from ..scheduler.base import BaseScheduler
from ..spider.base import BaseSpider


class SimpleSpider(BaseSpider):

    def __init__(self, scheduler: BaseScheduler = None):
        super(SimpleSpider, self).__init__(scheduler)

    def start(self):
        http_client = httpclient.AsyncHTTPClient()
        while not self.scheduler.empty():
            request = self.scheduler.get()
            http_client.fetch(request, self.parse)

    def parse(self, response: HTTPResponse):
        if response.error:
            yield ("Error:", response.error)
        else:
            yield ("Body:", response.body)



from typing import Generator
from tornado.ioloop import IOLoop
from ..scheduler.base import BaseScheduler
from tornado.httpclient import HTTPRequest, HTTPResponse
from .. import setting


class BaseSpider(object):
    """
    Interface for spider.
    """

    def __init__(self,
                 scheduler: BaseScheduler = None):
        self.scheduler = scheduler

    def start(self) -> Generator:
        """
        Spider start function. Call once on spider start.
        :return: a generator of tuple, HTTPRequest object and callback func
        """
        pass

    def parse(self, response: HTTPResponse) -> Generator:
        """
        Default parse function.
        todo: how to handle this function's request objects
        :param response:
        :return:
        """
        pass

    def run(self):
        """
        run spider
        :return: None
        """

        for item in self.start():
            if isinstance(item, HTTPRequest):
                req, cb = item, self.parse
            elif isinstance(item, tuple):
                assert isinstance(item[0], HTTPRequest) and isinstance(item[1], function)
                req, cb = item
            else:
                raise TypeError('start func return type error')
            self.scheduler.put(req, cb)

        loop = IOLoop.instance()
        try:
            loop.start()
        except KeyboardInterrupt:
            loop.stop()



from setuptools import setup, find_packages

setup(
    name='charlotte',
    packages=find_packages(exclude=('tests', 'tests.*')),
    version='0.0.1',
    url='https://github.com/Jack-Kingdom/hummingbird',
    description='Lightweight and expandable spider framework.',
    long_description=open('README.md').read(),

    author='Jack King',
    author_email='email@qiaohong.org',
    license='MIT',

    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Natural Language :: English',
    ], install_requires=[
        'tornado'
    ]
)
